@@ -1,3 +1,5 @@
+""" Implementation of class dendrite, which quries endpoints with tensors.
+"""
 # The MIT License (MIT)
 # Copyright © 2021 Yuma Rao
 
@@ -15,13 +17,17 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 # DEALINGS IN THE SOFTWARE.
 
-from bittensor._endpoint import endpoint
-import torch
+from types import SimpleNamespace
 from typing import Tuple, List, Union, Optional
-from torch.autograd.function import once_differentiable
-import bittensor
 
+import torch
+from torch.autograd.function import once_differentiable
 from loguru import logger
+
+import bittensor
+from bittensor._endpoint.endpoint_impl import Endpoint
+import bittensor.utils.stats as stat_utils
+
 logger = logger.opt(colors=True)
 
 # dummy tensor that triggers autograd 
@@ -29,6 +35,8 @@ DUMMY = torch.empty(0, requires_grad=True)
 
 # Helper function for filling nill (zero) responses on failures.
 def nill_response_for(inputs):
+    """ Get zero matrix with the same size as inputs
+    """
     if torch.numel(inputs) == 0:
         return torch.tensor([])
     return torch.zeros( (inputs.size(0), inputs.size(1), bittensor.__network_dim__), dtype=torch.float32)
@@ -61,10 +69,25 @@ class Dendrite( torch.autograd.Function ):
                 receptor_pool (:obj:`bittensor.ReceptorPool`, `required`):
                     bittensor receptor pool
         """
+        super().__init__()
         self.config = config
         self.wallet = wallet
         self.receptor_pool = receptor_pool
 
+        # ---- Dendrite stats
+        # num of time we have sent request to a peer, received successful respond, and the respond time
+        self.stats = SimpleNamespace(
+            requested_peers_count = {},
+            responded_peers_count = {},
+            peers_respond_time = {}
+        )
+
+    def __str__(self):
+        return "Dendrite({}, {})".format(self.wallet.hotkey.ss58_address, self.receptor_pool)
+
+    def __repr__(self):
+        return self.__str__()
+
     @staticmethod
     def forward(
             ctx, 
@@ -107,6 +130,9 @@ class Dendrite( torch.autograd.Function ):
             Returns:
                 codes (:obj:`torch.LongTensor` of shape :obj:`(n_endpoints)` `required`):
                     Return code associated with forward call.
+
+                times (:obj:`torch.FloatTensor` of shape :obj:`[ num_endpoints ]`, `required`):
+                    times per call.
                 
                 outputs (:obj:`List[torch.FloatTensor]` of shape :obj:`n_endpoints * (batch_size, sequence_len, bittensor.__network_dim__)`, `required`):
                         Output encodings of inputs produced by the remote endpoints. Non-responses are zeroes of common shape.
@@ -114,20 +140,21 @@ class Dendrite( torch.autograd.Function ):
         ctx.receptor_pool = dendrite.receptor_pool
         ctx.endpoints, ctx.inputs, ctx.modality, ctx.timeout, ctx.does_requires_grad = endpoints, inputs, modality, timeout, requires_grad
         inputs = [ x.cpu().clone().detach() for x in inputs ]
-        forward_outputs, forward_codes = ctx.receptor_pool.forward(
+        forward_outputs, forward_codes, forward_times = ctx.receptor_pool.forward(
             endpoints = endpoints, 
             inputs = inputs, 
             modality = modality,
             timeout = timeout
         )
         ctx.forward_codes = forward_codes
-        return (torch.tensor(forward_codes, dtype=torch.int64), *forward_outputs)
+        return (torch.tensor(forward_codes, dtype=torch.int64), torch.tensor(forward_times, dtype=torch.float32), *forward_outputs)
 
     @staticmethod
     @once_differentiable
     def backward( 
             ctx, 
             unused_code_grads: torch.FloatTensor,
+            unused_time_grads: torch.FloatTensor,
             *output_grads: torch.FloatTensor
         ) -> Tuple[ Optional[torch.Tensor], ... ]:
         """ Internal autograd-friendly Backward RPC call to a list of neuron endpoints.
@@ -139,6 +166,9 @@ class Dendrite( torch.autograd.Function ):
                 unused_code_grads: (:obj:`List[torch.Tensor]` of shape :obj:`(shape)`, `required`):
                     Gradients of this function's codes. (Unused)
 
+                unused_time_grads: (:obj:`List[torch.Tensor]` of shape :obj:`(shape)`, `required`):
+                    Gradients of this function's query times. (Unused)
+
                 grads (:obj:`List[torch.Tensor]` of shape :obj:`(shape)`, `required`):
                     Gradients of this function's outputs computed during the loss.backward() call.
             
@@ -150,7 +180,7 @@ class Dendrite( torch.autograd.Function ):
         """
         if ctx.does_requires_grad:
             grads_cpu = [ x.cpu().clone().detach() for x in output_grads ]
-            input_grads, _ =  ctx.receptor_pool.backward (
+            input_grads, _, _ =  ctx.receptor_pool.backward (
                 endpoints = ctx.endpoints, 
                 inputs_x = ctx.inputs, 
                 grads_dy = grads_cpu, 
@@ -171,46 +201,51 @@ class Dendrite( torch.autograd.Function ):
                 timeout: int = None,
                 requires_grad: bool = None
             ) -> Tuple[torch.LongTensor, List[torch.Tensor]]:
-            r""" Internal Forward tensor inputs to a list of neuron endpoints.
+        r""" Internal Forward tensor inputs to a list of neuron endpoints.
 
-                Args:
-                    endpoints (:obj:`List[bittensor.Endpoint]` of shape :obj:`(num_endpoints)`, `required`):
-                        List of remote endpoints which match length of inputs. Tensors from inputs are sent forward to these endpoints.
+            Args:
+                endpoints (:obj:`List[bittensor.Endpoint]` of shape :obj:`(num_endpoints)`, `required`):
+                    List of remote endpoints which match length of inputs. Tensors from inputs are sent forward to these endpoints.
 
-                    inputs (:obj:`List[torch.Tensor]` of shape :obj:`(num_endpoints * [shape])`, `required`):
-                        List of tensors to send to corresponsing endpoints. Tensors are of arbitrary type and shape depending on the
-                        modality.
+                inputs (:obj:`List[torch.Tensor]` of shape :obj:`(num_endpoints * [shape])`, `required`):
+                    List of tensors to send to corresponsing endpoints. Tensors are of arbitrary type and shape depending on the
+                    modality.
 
-                    modality (:obj:`bittensor.proto.Modality` of shape :obj:`(1)`, `required`):
-                        Bittensor forward modality type. Enum in [TEXT, IMAGE, TENSOR]
+                modality (:obj:`bittensor.proto.Modality` of shape :obj:`(1)`, `required`):
+                    Bittensor forward modality type. Enum in [TEXT, IMAGE, TENSOR]
 
-                    timeout (int, default = dendrite.timeout, `required`):
-                        request timeout.
+                timeout (int, default = dendrite.timeout, `required`):
+                    request timeout.
 
-                    requires_grad (int, default = dendrite.requires_grad, `optional`):
-                        If true, the backward pass triggers passing gradients on the wire.
+                requires_grad (int, default = dendrite.requires_grad, `optional`):
+                    If true, the backward pass triggers passing gradients on the wire.
 
-                Returns:
-                    codes (:obj:`List[torch.LongTensor]` of shape :obj:`[num_endpoints]`, `required`):
-                        dendrite call return codes.
+            Returns:
+                responses (:obj:`List[torch.FloatTensor]` of shape :obj:`(batch_size, sequence_len, bittensor.__network_dim__)`, `required`):
+                    Output encodings of inputs produced by the remote endpoints. Non-responses are zeroes of common shape.
 
-                    responses (:obj:`List[torch.FloatTensor]` of shape :obj:`(batch_size, sequence_len, bittensor.__network_dim__)`, `required`):
-                        Output encodings of inputs produced by the remote endpoints. Non-responses are zeroes of common shape.
-            """
-            timeout = timeout if timeout != None else self.config.dendrite.timeout 
-            requires_grad = requires_grad if requires_grad != None else self.config.dendrite.requires_grad 
-            forward_response = Dendrite.apply(
-                self,
-                DUMMY, 
-                endpoints, 
-                modality,
-                timeout,
-                requires_grad,
-                *inputs
-            )
-            codes = forward_response[0]
-            tensors = forward_response[1:]
-            return tensors, codes
+                codes (:obj:`List[torch.LongTensor]` of shape :obj:`[num_endpoints]`, `required`):
+                    dendrite call return codes.
+
+                times (:obj:`torch.FloatTensor` of shape :obj:`[ num_endpoints ]`, `required`):
+                    times per call.
+
+        """
+        timeout = timeout if timeout != None else self.config.dendrite.timeout 
+        requires_grad = requires_grad if requires_grad != None else self.config.dendrite.requires_grad 
+        forward_response = Dendrite.apply(
+            self,
+            DUMMY, 
+            endpoints, 
+            modality,
+            timeout,
+            requires_grad,
+            *inputs
+        )
+        codes = forward_response[0]
+        times = forward_response[1]
+        tensors = forward_response[2:]
+        return tensors, codes, times
 
     def forward_image(
         self, 
@@ -218,7 +253,7 @@ class Dendrite( torch.autograd.Function ):
         inputs: List[ torch.FloatTensor ],
         timeout: int = None,
         requires_grad: bool = None
-    ) -> Tuple[ Union[List[torch.FloatTensor], torch.FloatTensor], torch.LongTensor]:
+    ) -> Tuple[ Union[List[torch.FloatTensor], torch.FloatTensor], torch.LongTensor, torch.FloatTensor]:
         r""" Forward image inputs to endpoints.
 
           Args:
@@ -241,9 +276,12 @@ class Dendrite( torch.autograd.Function ):
 
                 codes (:obj:`torch.LongTensor` of shape :obj:`[ num_endpoints ]`, `required`):
                     dendrite call return ops.
+
+                times (:obj:`torch.FloatTensor` of shape :obj:`[ num_endpoints ]`, `required`):
+                    times per call.
         """
          # Check types.
-        if not isinstance(endpoints, list) and not isinstance(endpoints, bittensor._endpoint.endpoint_impl.Endpoint):
+        if not isinstance(endpoints, list) and not isinstance(endpoints, Endpoint):
             raise ValueError('endpoints must be of type list or bittensor.Endpoint. Got {}'.format(type(endpoints)))
 
         if not isinstance(inputs, list) and not isinstance(inputs, torch.FloatTensor):
@@ -275,7 +313,7 @@ class Dendrite( torch.autograd.Function ):
         # Check list types.
         if not isinstance(inputs[0], torch.FloatTensor):
             raise ValueError('inputs must be of type torch.FloatTensor. Got {}'.format(type(inputs[0])))
-        if not isinstance(endpoints[0], bittensor._endpoint.endpoint_impl.Endpoint):
+        if not isinstance(endpoints[0], Endpoint):
             raise ValueError('endpoints must be of type bittensor.Endpoint. Got {}'.format(type(endpoints)))
 
         # Check shape.
@@ -284,7 +322,7 @@ class Dendrite( torch.autograd.Function ):
             raise ValueError(error_msg)
         
         # Make calls.
-        responses, codes = self._forward(
+        responses, codes, times = self._forward(
             endpoints = endpoints, 
             inputs = inputs, 
             modality = bittensor.proto.Modality.IMAGE,
@@ -297,7 +335,8 @@ class Dendrite( torch.autograd.Function ):
             responses = responses[0]
 
         # Return.
-        return responses, codes
+        self.update_stat(endpoints, codes, times)
+        return responses, codes, times
 
     def forward_tensor(
             self, 
@@ -305,7 +344,7 @@ class Dendrite( torch.autograd.Function ):
             inputs: List[ torch.FloatTensor ],
             timeout: int = None,
             requires_grad: bool = None
-        ) -> Tuple[ Union[List[torch.FloatTensor], torch.FloatTensor], torch.LongTensor]:
+        ) -> Tuple[ Union[List[torch.FloatTensor], torch.FloatTensor], torch.LongTensor, torch.FloatTensor]:
         r""" Forward tensor inputs to endpoints.
 
             Args:
@@ -328,9 +367,12 @@ class Dendrite( torch.autograd.Function ):
 
                 codes (:obj:`torch.LongTensor` of shape :obj:`[ num_endpoints ]`, `required`):
                     dendrite call return ops.
+
+                times (:obj:`torch.FloatTensor` of shape :obj:`[ num_endpoints ]`, `required`):
+                    times per call.
         """
         # Check types.
-        if not isinstance(endpoints, list) and not isinstance(endpoints, bittensor._endpoint.endpoint_impl.Endpoint):
+        if not isinstance(endpoints, list) and not isinstance(endpoints, Endpoint):
             raise ValueError('endpoints must be of type list or bittensor.Endpoint. Got {}'.format(type(endpoints)))
 
         if not isinstance(inputs, list) and not isinstance(inputs, torch.FloatTensor):
@@ -362,7 +404,7 @@ class Dendrite( torch.autograd.Function ):
         # Check list types.
         if not isinstance(inputs[0], torch.FloatTensor):
             raise ValueError('inputs must be of type torch.FloatTensor. Got {}'.format(type(inputs[0])))
-        if not isinstance(endpoints[0], bittensor._endpoint.endpoint_impl.Endpoint):
+        if not isinstance(endpoints[0], Endpoint):
             raise ValueError('endpoints must be of type bittensor.Endpoint. Got {}'.format(type(endpoints)))
 
         # Check shape.
@@ -374,7 +416,7 @@ class Dendrite( torch.autograd.Function ):
             raise ValueError(error_msg)
 
         # Make calls.
-        responses, codes = self._forward(
+        responses, codes, times = self._forward(
             endpoints = endpoints, 
             inputs = inputs, 
             modality = bittensor.proto.Modality.TENSOR,
@@ -387,8 +429,8 @@ class Dendrite( torch.autograd.Function ):
             responses = responses[0]
 
         # Return.
-        return responses, codes
-
+        self.update_stat(endpoints, codes, times)
+        return responses, codes, times
 
     def forward_text(
             self,
@@ -396,7 +438,7 @@ class Dendrite( torch.autograd.Function ):
             inputs: Union[str, List[str], List[torch.LongTensor], torch.LongTensor],
             timeout: int = None,
             requires_grad: bool = None
-        ) -> Tuple[ Union[List[torch.FloatTensor], torch.FloatTensor], torch.LongTensor]:
+        ) -> Tuple[ Union[List[torch.FloatTensor], torch.FloatTensor], torch.LongTensor, torch.FloatTensor]:
         r""" Forward text inputs to a list of neuron endpoints and block until responses or timeout.
 
                 Args:
@@ -430,30 +472,32 @@ class Dendrite( torch.autograd.Function ):
                     codes (:obj:`torch.LongTensor` of shape :obj:`[ num_endpoints ]`, `required`):
                         dendrite call return ops.
 
+                    times (:obj:`torch.FloatTensor` of shape :obj:`[ num_endpoints ]`, `required`):
+                        times per call.
         """
         # To be filled. Inputs and endpoint must be list with the same number of elements.
         formatted_inputs = []
         formatted_endpoints = []
 
         # <<Helper function>> optional casts and then checks shape of inputs.
-        def cast_and_check_tensor_input( input ) -> torch.LongTensor:
-            if not isinstance ( input, torch.LongTensor):
+        def cast_and_check_tensor_input( tensor_input ) -> torch.LongTensor:
+            if not isinstance ( tensor_input, torch.LongTensor):
                 try:
-                    input = input.to( torch.int64 )
+                    tensor_input = tensor_input.to( torch.int64 )
                 except Exception as E:
-                    error_msg = 'Error while casting tensor input {} to int64 {}'.format(input, E)
-                    raise ValueError(error_msg)
-            if not ( isinstance(input, torch.cuda.LongTensor) or isinstance(input, torch.LongTensor)) :
-                raise ValueError('input {} must be of type torch.LongTensor. Got {}'.format(input, type(input)))
+                    error_msg = 'Error while casting tensor input {} to int64 {}'.format(tensor_input, E)
+                    raise ValueError(error_msg) from ValueError()
+            if not ( isinstance(tensor_input, torch.cuda.LongTensor) or isinstance(tensor_input, torch.LongTensor)) :
+                raise ValueError('input {} must be of type torch.LongTensor. Got {}'.format(tensor_input, type(tensor_input)))
             # Expand shape if it is a singlular dimension.
-            if len( input.shape ) == 1:
-                input = input.view(1, -1)
+            if len( tensor_input.shape ) == 1:
+                tensor_input = tensor_input.view(1, -1)
 
             # Check shape.
-            if len( input.shape ) != 2:
+            if len( tensor_input.shape ) != 2:
                 error_msg = 'Text inputs should be rank 2 with semantic shape: [batch_size, sequence_len]'
                 raise ValueError(error_msg)
-            return input 
+            return tensor_input 
 
         # ---- Endpoints is singular.
         if isinstance( endpoints, bittensor.Endpoint ):
@@ -494,14 +538,11 @@ class Dendrite( torch.autograd.Function ):
                         Got {} """.format(endpoints)
             raise ValueError(error_msg)
                         
-
-
-
         # ---- Inputs is a string
         if isinstance( inputs, str ):
             # Encode to tensors.
             tokenizer = bittensor.tokenizer()
-            inputs_list = tokenizer.encode( inputs )
+            inputs_list = tokenizer( inputs )['input_ids']
             inputs_tensor = cast_and_check_tensor_input ( torch.tensor( [inputs_list], dtype=torch.int64 ) )
             # Expand to length.
             formatted_inputs = [ inputs_tensor for _ in formatted_endpoints ]
@@ -544,7 +585,7 @@ class Dendrite( torch.autograd.Function ):
             raise ValueError(error_msg)
 
         # Make calls.
-        responses, codes = self._forward(
+        responses, codes, times = self._forward(
             endpoints = formatted_endpoints, 
             inputs = formatted_inputs, 
             modality = bittensor.proto.Modality.TEXT,
@@ -553,4 +594,63 @@ class Dendrite( torch.autograd.Function ):
         )
 
         # Return.
-        return responses, codes
\ No newline at end of file
+        self.update_stat(formatted_endpoints, codes, times)
+        return responses, codes, times
+
+    def update_stat(self, endpoints, return_ops, query_times):
+        r""" Update dendrite stat according to the response we get from peers.
+        Updates were saved to self.stats.
+            Args:
+                endpoints (:obj:`List[bittensor.Endpoint]` of shape :obj:`(num_endpoints)`, `required`):
+                    The set of endpoints that dendrite sent request to.
+
+                return_ops (:obj:`torch.LongTensor` of shape :obj:`[ num_endpoints ]`, `required`):
+                    Dendrite call return ops.
+
+                query_times (:obj:`torch.FloatTensor` of shape :obj:`[ num_endpoints ]`, `required`):
+                    Times per call.
+        """
+        # ---- uids that we have sent request to.
+        uids = torch.tensor([e.uid for e in endpoints])
+
+        # ---- uids that gave us successful respond.
+        success_ids= torch.where( return_ops == bittensor.proto.ReturnCode.Success )[0]
+        
+        # ---- For each uid, check we have a stats column for this peer and aggregate to stats.
+        for uid, time in zip(uids, query_times):
+            if uid in self.stats.requested_peers_count.keys():
+                self.stats.requested_peers_count[uid].update(1)
+                self.stats.peers_respond_time[uid].update(time)
+
+            else:
+                self.stats.requested_peers_count[uid] = stat_utils.timed_rolling_avg(1, 0.01)
+                self.stats.responded_peers_count[uid] = stat_utils.timed_rolling_avg(1, 0.01)
+                self.stats.peers_respond_time[uid] = stat_utils.timed_rolling_avg(time, 0.01)
+
+        
+        # --- Aggregating result to stats 
+        for uid in uids[success_ids]:
+            if uid in self.stats.requested_peers_count.keys():
+                self.stats.responded_peers_count[uid].update(1)
+            else:
+                self.stats.responded_peers_count[uid] = stat_utils.timed_rolling_avg(1, 0.01)
+
+    def to_wandb(self):
+        r""" Return a dictionary of axon stat for wandb logging
+            
+            Return:
+                wandb_info (:obj:`Dict`)
+        """
+        wandb_info = {}
+
+        # ---- Dendrite stats per pubkey for wandb 
+        for uid in self.stats.requested_peers_count.keys():
+            respond_rate = self.stats.responded_peers_count[uid].value / self.stats.requested_peers_count[uid].value
+           
+            uid_str = str(uid.item()).zfill(3)
+            wandb_info[f'dend_quested uid: {uid_str}']= self.stats.requested_peers_count[uid].value
+            wandb_info[f'dend_responded uid: {uid_str}']= self.stats.responded_peers_count[uid].value
+            wandb_info[f'dend_respond_time uid: {uid_str}']= self.stats.peers_respond_time[uid].value
+            wandb_info[f'dend_respond_rate uid: {uid_str}']= respond_rate
+
+        return wandb_info
\ No newline at end of file
